/*
 * Copyright (c) 2005 Björn Steinbrink
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <syslog.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "commands.h"

#define SOCKET_PATH "/tmp/vservd.sock"

int s1;

int main(int argc, char **argv)
{
	int tty;
	struct winsize ws;
	struct sockaddr_un remote;
	int len, done = 0;
	struct termios tios, old;

	tty = open("/proc/self/fd/0", 0, 0);
	ioctl(tty, TIOCGWINSZ, &ws);

	if(tcgetattr(tty, &old))
		perror("get1");

	s1 = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s1 == -1) {
		perror("Couldn't create socket");
		exit(EXIT_FAILURE);
	}

	remote.sun_family = AF_UNIX;
	strcpy(remote.sun_path, SOCKET_PATH);
	len = strlen(remote.sun_path) + sizeof(remote.sun_family);

	if (connect(s1, (struct sockaddr*)&remote, len) == -1) {
		perror("Can't connect to socket");
		exit(EXIT_FAILURE);
	}

	printf("> ");
	fflush(0);
	while (!done) {
		fd_set fds;
		struct msghdr msg;
		struct iovec iov;

		FD_ZERO(&fds);
		FD_SET(s1, &fds);
		FD_SET(0, &fds);

		select(s1+1, &fds, NULL, NULL, NULL);

		if (FD_ISSET(s1, &fds)) {
			int t, master = 0;
			char answer[255];
			char control[1024];
			struct cmsghdr *cmsg;

			memset(&msg, 0, sizeof(msg));
			iov.iov_base = answer;
			iov.iov_len = sizeof(answer)-1;
			msg.msg_iov = &iov;
			msg.msg_iovlen = 1;
			msg.msg_control = control;
			msg.msg_controllen = sizeof(control);

			t = recvmsg(s1, &msg, 0);

			if (t == -1) {
				perror("recvmsg");
				exit(EXIT_FAILURE);
			} else if (t == 0) {
				printf("\nServer closed connection\n");
				done = 1;
				continue;
			}

			if (!strncmp(answer, "MASTER", 6)) {
				/* Send through terminal data */
				cmsg = CMSG_FIRSTHDR(&msg);
				while (cmsg != NULL) {
					if (cmsg->cmsg_level == SOL_SOCKET
							&& cmsg->cmsg_type  == SCM_RIGHTS) {
						master =  *(int *) CMSG_DATA(cmsg);
						break;
					}
					cmsg = CMSG_NXTHDR(&msg, cmsg);
				}
				if (!master)
				{
					perror("No master");
					break;
				}
				if(fcntl(master, F_SETFL, O_NONBLOCK))
					perror("fcntl");
				if(fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK))
					perror("fcntl");
				if(fcntl(STDOUT_FILENO, F_SETFL, O_NONBLOCK))
					perror("fcntl");
				if(ioctl(tty, TIOCGWINSZ, &ws))
					perror("ioctl");
				if(ioctl(master, TIOCSWINSZ, &ws))
					perror("ioctl");

				int done=0;
				while (!done) {
					fd_set fds;
					FD_ZERO(&fds);
					FD_SET(STDIN_FILENO, &fds);
					FD_SET(master, &fds);

					select(master+1, &fds, NULL, NULL, NULL);

					if(tcgetattr(master, &tios))
						perror("get");
					cfmakeraw(&tios);
					if(tcsetattr(tty, TCSANOW, &tios))
						perror("set");

					if (FD_ISSET(master, &fds)) {
						char buffer[4096];
						size_t bytes;
						while ((bytes = read(master, buffer, 4096)) != -1) {
							int offset;
							int n;
							offset = 0;
							if (!bytes)
								done = 1;
							while(bytes) {
								n = write(STDOUT_FILENO, buffer+offset, bytes);
								if (n == -1 && errno != EAGAIN) {
									perror("write");
									tcsetattr(tty, TCSANOW, &old);
									exit(EXIT_FAILURE);
									break;
								}
								if (n == -1) n = 0;
								offset += n;
								bytes -= n;
							}
						}


						if (errno != EAGAIN) {
							done = 1;
						}
					}

					if (FD_ISSET(STDIN_FILENO, &fds)) {
						char buffer[4096];
						int bytes;
						bytes = read(STDIN_FILENO, buffer, 4096);
						if (!bytes)
							done = 1;
						write(master, buffer, bytes);
					}
				}
				tcsetattr(tty, TCSANOW, &old);
				printf("\nfinished\n> ");
				fflush(0);
			}
		}

		if (FD_ISSET(0, &fds)) {
			char input[255];
			fgets(input, 255, stdin);
			if (feof(stdin)) {
				printf("exit\n");
				done = 1;
			} else {
				struct vservd_cmd cmd;

				printf("> ");
				fflush(0);

				memset(&msg, 0, sizeof(msg));
				iov.iov_base = &cmd;
				iov.iov_len = sizeof(cmd);
				msg.msg_iov = &iov;
				msg.msg_iovlen = 1;

				if (!strncmp(input, "quit", 4)) {
					cmd.type = CMD_QUIT;
				} else if (!strncmp(input, "create ", 7)) {
					cmd.type = CMD_CREATE;
					sscanf(input, "create %d", &cmd.xid);
				} else if (!strncmp(input, "enter ", 6)) {
					cmd.type = CMD_ENTER;
					sscanf(input, "enter %d", &cmd.xid);
				} else
					continue;

				if (sendmsg(s1, &msg, 0) == -1) {
					perror("sendmsg");
					exit(EXIT_FAILURE);
				}
			}
		}
	}

	close(s1);

	exit(EXIT_SUCCESS);
}
